#include "chull.h"
#include <assert.h>
#include <iostream>

inline bool CHull::hasnil(const Point *p, const Point *q, const Point *r)
{
    return (p == Point::NIL || q == Point::NIL || r == Point::NIL);
}
inline double CHull::turn(const Point *p, const Point *q, const Point *r)
{
    return (q->x() - p->x()) * (r->y() - p->y()) -
            (r->x() - p->x()) * (q->y() - p->y());
}
inline double CHull::time(const Point *p, const Point *q, const Point *r)
{
    if (p == Point::NIL || q == Point::NIL || r == Point::NIL)
        return Point::INF;
    return ((q->x() - p->x()) * (r->z() - p->z()) -
            (r->x() - p->x()) * (q->z() - p->z())) / turn(p, q, r);
}

CHull::CHull(PointSet *set)
{
    assert(set != 0);
    _set = set;
    _set->ref();

    _A = new Point*[2 * _set->size()];
    _B = new Point*[2 * _set->size()];

    _running = false;
}

CHull::~CHull()
{
    // if you delete it while a thread is running, that would be bad.
    assert(_running == false);

    delete _A;
    delete _B;
    _set->unref();
}

void CHull::hull(bool bottom, Point *list, int n, Point **A, Point **B, int level)
{
    Point *u, *v, *mid;
    double oldt;
    int i, j, k;

    // base case is a single point
    if (n == 1) {
        // remove the point from the list
        list->next = Point::NIL;
        list->prev = Point::NIL;
        // return no faces for the hull
        A[0] = Point::NIL;
        return;
    }

    // find the middle of the current problem set
    for (u = list, i = 0; i < n/2-1; u = u->next, ++i);
    mid = v = u->next;

    // recurse on each side of the middle, swapping the A and B arrays
    hull(bottom, list, n/2, B, A, level+1);       // build left side
    hull(bottom, mid, n-n/2, B+n/2*2, A+n/2*2, level+1);  // build right side

    // stop before doing merge
#if 0
    pauseit();
#endif
    if (_abort) return;

    // merge process

    // find the initial bridge
    for (;;) {
        if (bottom) {
            // start in the middle and move the bridge vertices out to each side
            if (!hasnil(u, v, v->next) && turn(u, v, v->next) < 1.0) v = v->next;
            else if (!hasnil(u->prev, u, v) && turn(u->prev, u, v) < 1.0) u = u->prev;
            else break;
        }
        else {
            // start in the middle and move the bridge vertices out to each side
            if (!hasnil(u, v, v->next) && turn(u, v, v->next) > 1.0) v = v->next;
            else if (!hasnil(u->prev, u, v) && turn(u->prev, u, v) > 1.0) u = u->prev;
            else break;
        }
    }

    if (level == 0) {
        if (!_abort) emit initialHull(bottom, level, u, v);
        //pauseit();
    }

    // merge by tracking bridge uv over time
    i = 0;              // offset into B for the left side hull
    j = n/2*2;          // offset into B for the right side hull
    k = 0;              // offset into A where we are writing
    oldt = -Point::INF; // start at time -infinity
    while (true) {
        int l, minl;
        double t, newt;
        Point *p[6], *q[6], *r[6]; // a face of the CH, q is the vertex for which the time event occurs
        bool event = false;

        // find the next possible events
        p[0] = B[i]->prev; q[0] = B[i];    r[0] = B[i]->next; // left hull
        p[1] = B[j]->prev; q[1] = B[j];    r[1] = B[j]->next; // right hull
        p[2] = u;          q[2] = u->next; r[2] = v;          // bridge..
        p[3] = u->prev;    q[3] = u;       r[3] = v;
        p[4] = u;          q[4] = v->prev; r[4] = v;
        p[5] = u;          q[5] = v;       r[5] = v->next;

        // find an event that is past the current time
        for (newt = Point::INF, l = 0; l < 6; ++l) {
            t = time(p[l], q[l], r[l]);
            if (t > oldt && t < newt) {
                // found an event in the future, closer than any others found (newt)
                minl = l;
                newt = t;
            }
        }
        if (newt == Point::INF) {
            // no events found, we have the whole merged hull
            break;
        }

        switch (minl) {
        case 0: // left side
            event = B[i]->x() < u->x(); // left of the bridge ? (not hidden under the bridge)
            break;
        case 1: // right side
            event = B[j]->x() > v->x(); // right of the bridge ? (not hidden under the bridge)
            break;
        default:
            // the other cases are always a change to the 2d hull
            event = true;
        }

        if (event && level == 0) {
            // tell the visuals that we found a face of the 3d hull. i.e. the points
            // p, q, r are consecutive points on the 2d hull and became colinear
            if (!_abort) emit nextEvent(newt, p[minl], q[minl], r[minl]);
            // then pause before changing the 2d projected hull, to give the
            // visuals a chance to react before the 2d hull changes
            pauseit();
        }

        // handle the next event which gives a face of the merged hull in A[k]
        // and update the 2d hull for the event
        switch (minl) {
        case 0: // left side
            if (event) A[k++] = B[i];
            B[i++]->addremove();
            break;
        case 1: // right side
            if (event) A[k++] = B[j];
            B[j++]->addremove();
            break;
        case 2: // bridge
            u = u->next;
            A[k++] = u;
            break;
        case 3: // bridge
            A[k++] = u;
            u = u->prev;
            break;
        case 4: // bridge
            v = v->prev;
            A[k++] = v;
            break;
        case 5: // bridge
            A[k++] = v;
            v = v->next;
            break;
        }

        if (event && level == 0) {
            // notify the visuals that the new 2d projected hull is available
            if (!_abort) emit updateHull(u, v);
            // wait for the visuals before going on to find the next event
            pauseit();
        }

        oldt = newt;
    }

    A[k] = Point::NIL; // mark the end of the merged hull

    if (level == 0) {
        // tell the visuals that we're done merging
        if (!_abort) emit doneLevel(bottom, level);
    }

    // now go back in time to update pointers
    // the vertices that are on the merged 2d hull at t=inf are
    // not connected anymore to their neighbours when the event
    // occured where they were added/removed to the 2d hull, so
    // go back and reconnect them to those neighbours so that
    // each A[k] can be used as a face, along with A[k].prev and
    // A[k] next

    // obviously, this isn't important for the last time through
    // unless you plan to use this array to spit out a list of
    // the faces, and it breaks the 2d projection view so we'll
    // just skip it for the last level

    if (level == 0) return;

    u->next = v; // make the bridge end points..
    v->prev = u; // ..next to each other
    for (--k; k >= 0; --k) { // from back of A[k] to front
        if (A[k]->x() <= u->x() || A[k]->x() >= v->x()) {
            // on the left or right hull

            A[k]->addremove();
            if (A[k] == u)
                u = u->prev;
            else if (A[k] == v)
                v = v->next;
        }
        else {
            // inside the bridge, so it was a bridge endpoint

            // put it between the current two bridge endpoints, u and v
            u->next = A[k];
            A[k]->prev = u;
            v->prev = A[k];
            A[k]->next = v;

            // make it a bridge endpoint, on the left or right side
            if (A[k]->x() < mid->x())
                u = A[k];  // on the left side
            else
                v = A[k];  // on the right side
        }

        if (level == 0) {
            if (!_abort) emit nextEvent(Point::INF, A[k]->prev, A[k], A[k]->next);
            pauseit();
        }
    }
}

void CHull::pauseit()
{
    if (!_abort) {
        //std::cout << "pauseit (" << _pause_count.available() << " next() in the queue)\n";
        //std::cout.flush();
        _pause_count.acquire(1);
    }
}

void CHull::next()
{
    _pause_count.release(1);
    //std::cout << "next (" << _pause_count.available() << " next() in the queue)\n";
    //std::cout.flush();
}

void CHull::abort()
{
    _abort = true;
    // make sure it's not in a lock.  if it was between the check for
    // _abort and the lock then this will make sure it's not stuck
    _pause_count.release(1);
    //std::cout << "abort (" << _pause_count.available() << " next() in the queue)\n";
    //std::cout.flush();
}

void CHull::restart()
{
    if (!_running)
        start(); // the thread isn't running, so just make it start going again
    else {
        _abort = _restart = true;
        // make sure it's not in a lock.  if it was between the check for
        // _abort and the lock then this will make sure it's not stuck
        _pause_count.release(1);

        //std::cout << "abort (" << _pause_count.available() << " next() in the queue)\n";
        //std::cout.flush();
    }
}

void CHull::run()
{
    _running = true;

    do {
        // clear any next() calls already made
        _pause_count.acquire(_pause_count.available());
        
        _abort = false;
        _restart = false;

        pauseit(); // wait for next() to start
        hull(true, _set->sort(), _set->size(), _A, _B, 0);

        pauseit(); // wait for next() to start
        hull(false, _set->sort(), _set->size(), _A, _B, 0);
    } while (_restart);

    emit done(_abort);

    _running = false;
}
