#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "pointset.h"
#include "chull.h"
#include <QInputDialog>
#include <QFileDialog>
#include <QMessageBox>
#include <iostream>

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), ui(new Ui::MainWindowClass), _set(0), _hull(0),
    _waiting_for_anim(false)
{
    ui->setupUi(this);
    connect(ui->actionMinimize, SIGNAL(triggered()), this, SLOT(showMinimized()));
    connect(ui->actionZoom, SIGNAL(triggered()), this, SLOT(toggleMaximized()));
    connect(ui->mergeView, SIGNAL(showingTime(double)), this, SLOT(projectionShowingTime(double)));
    this->setFocus();  // focus the main window (i.e. "nothing") by default

    // set up the gui stuff to have the same value as the internals
    ui->mergeView->setAnimationSpeed(ProjectionDisplay::DEFAULT_ANIM_SPEED_MS);
    ui->animationSpeed->setValue(ProjectionDisplay::DEFAULT_ANIM_SPEED_MS);
    ui->autoZoom->setChecked(ui->mergeView->getAutoZoom());

    // show a default point set
    this->on_actionUse_Default_Point_Set_triggered();
}

MainWindow::~MainWindow()
{
    setPointSet(0);

    delete ui;
}

void MainWindow::toggleMaximized()
{
    if (isMaximized()) showNormal();
    else showMaximized();
}

void MainWindow::on_actionNew_Point_Set_triggered()
{
    QInputDialog in(this);
    int ret;

    in.setModal(true);
    in.setLabelText("Enter the size of the point set:");
    in.setIntMinimum(3);
    in.setIntMaximum(1000);
    in.setIntValue(8);
    in.setWindowTitle("New Point Set");
    in.setInputMode(QInputDialog::IntInput);
    ret = in.exec();
    if (ret)
        setPointSet(PointSet::generate(in.intValue()));
        //setPointSet(PointSet::generateDefault());
}

void MainWindow::on_actionReset_3d_View_triggered()
{
    ui->pointsView->resetView();
}

void MainWindow::setPointSet(PointSet *set)
{
    if (_hull) {
        disconnect(_hull, SIGNAL(updateHull(const Point*,const Point*)),
                   this, SLOT(updateHull(const Point*,const Point*)));
        disconnect(_hull, SIGNAL(doneLevel(bool, int)),
                   this, SLOT(doneLevel(bool, int)));
        disconnect(_hull, SIGNAL(initialHull(bool,int,const Point*,const Point*)),
                   this, SLOT(initialHull(bool,int,const Point*,const Point*)));
        disconnect(_hull, SIGNAL(nextEvent(double,const Point*,const Point*,const Point*)),
                   this, SLOT(nextEvent(double,const Point*,const Point*,const Point*)));
        _hull->abort();
        _hull->wait();
    }
    delete _hull;
    _hull = 0;

    if (_set) _set->unref();

    _set = set;
    ui->pointsView->setPointSet(set);
    ui->mergeView->setPointSet(set);

    if (set) {
        _hull = new CHull(set);

        connect(_hull, SIGNAL(updateHull(const Point*,const Point*)),
                this, SLOT(updateHull(const Point*,const Point*)),
                Qt::BlockingQueuedConnection);
        connect(_hull, SIGNAL(doneLevel(bool, int)),
                this, SLOT(doneLevel(bool, int)),
                Qt::BlockingQueuedConnection);
        connect(_hull, SIGNAL(initialHull(bool,int,const Point*,const Point*)),
                this, SLOT(initialHull(bool,int,const Point*,const Point*)),
                Qt::BlockingQueuedConnection);
        connect(_hull, SIGNAL(nextEvent(double,const Point*,const Point*,const Point*)),
                this, SLOT(nextEvent(double,const Point*,const Point*,const Point*)),
                Qt::BlockingQueuedConnection);

        // start computing the convex hull in another thread
        _hull->start();
    }
}

void MainWindow::updateHull(const Point *u, const Point *v)
{
    ui->mergeView->showHull(u, v);
}

void MainWindow::nextEvent(double t, const Point *p, const Point *q, const Point *r)
{
    ui->pointsView->highlightLastFace(true);
    ui->pointsView->addFace(p, q, r);
    ui->mergeView->showTime(t);
    ui->mergeView->highlightPoint(q);

    // when the mergeview is done showing this time, it'll throw a signal back to us
    _waitmutex.lock();
    _waiting_for_anim = true;
    _waitmutex.unlock();
}

void MainWindow::projectionShowingTime(double t)
{
    (void)t;
    animationFinished();
}

void MainWindow::animationFinished()
{
    _waitmutex.lock();
    if (_waiting_for_anim) {
        // the CHull pauses when it tells us about a new time, and we wait for the animation
        // before telling it to proceed. so when the projection is done animating, then
        // tell the CHull to proceed (or if we don't want to wait for it to finish that
        // animation
        _hull->next();
        _waiting_for_anim = false;
    }
    _waitmutex.unlock();
}

void MainWindow::initialHull(bool bottom, int level, const Point *u, const Point *v)
{
    (void)bottom;(void)level;

    // reset the 2d projection, and show the initial 2d projected hull
    ui->mergeView->reset();
    ui->mergeView->showHull(u, v);
}

void MainWindow::doneLevel(bool bottom, int level)
{
    (void)level;

    QString s;
    s.append("Done ");
    s.append(bottom ? "Bottom" : "Top");
    s.append(" Hull");
    ui->pointsView->showOSDText(s);

    // finish the projection and don't highlight any point
    ui->mergeView->showTime(Point::INF);
    ui->mergeView->highlightPoint(0);

    // don't highlight anything anymore
    ui->pointsView->highlightLastFace(false);
}

void MainWindow::on_actionNext_Hull_triggered()
{
    animationFinished(); // stop waiting for any animation, so we can go on to find the next face
    _hull->next();
}

void MainWindow::on_animationSpeed_valueChanged(int ms)
{
    ui->mergeView->setAnimationSpeed(ms);
}

void MainWindow::on_actionRestart_triggered()
{
    _hull->restart();
    
    // clear the visualizations
    ui->pointsView->clearFaces();
    ui->mergeView->reset();

    _waiting_for_anim = false;
}

void MainWindow::on_autoZoom_clicked(bool checked)
{
    ui->mergeView->setAutoZoom(checked);
}

void MainWindow::on_actionOpen_Point_Set_triggered()
{
    QString fname, *error = 0;
    PointSet *set;

    fname = QFileDialog::getOpenFileName(this, "Open Point Set...", QDir::homePath());
    if (fname.isNull()) return;

    if ((set = PointSet::load(fname, &error)))
        setPointSet(set);
    else {
        QMessageBox::critical(this, "Error", *error);
        delete error;
    }
}

void MainWindow::on_actionUse_Default_Point_Set_triggered()
{
    setPointSet(PointSet::generateDefault());
}

void MainWindow::on_actionAbout_triggered()
{
    const char *text =
            "Visualization of Chan's 3d Convex Hull Algorithm\n"
            "\n"
            "By Dana Jansens\n"
            "\n"
            "COMP 5409 Final Project\n"
            "Carleton University\n"
            "\n"
            "See the README file for more information";

    QMessageBox::about(this, "About 3dHull", text);
}
